.noselect {
  -webkit-touch-callout: none;
  /* iOS Safari */
  -webkit-user-select: none;
  /* Safari */
  -khtml-user-select: none;
  /* Konqueror HTML */
  -moz-user-select: none;
  /* Old versions of Firefox */
  -ms-user-select: none;
  /* Internet Explorer/Edge */
  user-select: none;
  /* Non-prefixed version, currently supported by Chrome, Edge, Opera and Firefox */
}

.clickable:hover,
.clickable *:hover {
  cursor: pointer;
}

.draggable,
.draggable:hover {
  cursor: move;
}

:root {
  /* most colors come from https://materialui.co/colors/ */
  --font-light: #fafafa;
  --font-dark: #212121;
  --alert-text: #bf360c;

  --background-light: #fafafa;
  --body-background: #e5e4e2;

  --box-blue: #e3f2fd;

  --header-font: var(--font-light);
  --header-background: #363636;
  --footer-font: #9e9e9e;
  --footer-link-font: #f5f5f5;
  --footer-background: #363636;

  --progress-active: #363636;
  --progress-inactive: #757575;

  --vote-nav-background: #757575;
  --review-ballot-background: var(--box-blue);

  --button-default: #bdbdbd;
  --button-default-hover: #9e9e9e;
  --button-default-active: #757575;
  --button-white: #fafafa;
  --button-white-hover: #f5f5f5;
  --button-white-active: #eeeeee;
  --button-blue: #1976d2;
  --button-blue-hover: #1565c0;
  --button-blue-active: #0d47a1;
  --button-grey: #757575;
  --button-grey-hover: #616161;
  --button-grey-active: #424242;

  --candidate-background: #eeeeee;
  --candidate-background-hover: #e0e0e0;
  --candidate-background-checked: #bbdefb;
  --candidate-background-checked-hover: #90caf9;
  --candidate-checkbox-background: #fafafa;
  --candidate-checkbox-border: #bdbdbd;
  --candidate-checkbox-checked: #1976d2;

  --page-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2),
    0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12);
  --box-shadow: 0px 2px 1px -1px rgba(0, 0, 0, 0.2),
    0px 1px 1px 0px rgba(0, 0, 0, 0.14), 0px 1px 3px 0px rgba(0, 0, 0, 0.12);
  --button-shadow: 0px 3px 1px -2px rgba(0, 0, 0, 0.2),
    0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12);
  --button-shadow-hover: 0px 2px 4px -1px rgba(0, 0, 0, 0.2),
    0px 4px 5px 0px rgba(0, 0, 0, 0.14), 0px 1px 10px 0px rgba(0, 0, 0, 0.12);
}
