.main-zone {
    flex: 4;
    margin-top: 10px;
    margin-bottom: 10px;
    margin-left: 20px;
    margin-right: 10px;
    min-height: 200px;

}

/* Pictures for delete / insert */

div.ins_sym {
    width: 1.8ex;
    height: 1.8ex;
    background-image: url("data:image/svg+xml;utf8,%3Csvg viewBox='0 0 10 10' xmlns='http://www.w3.org/2000/svg'%3E%3Ccircle cx='5' cy='5' r='4.5' stroke='black' fill='none'/%3E%3Cpath d='M 5 2 V 8 M 2 5 H 8' stroke='black'/%3E%3C/svg%3E");
    background-size: cover;
}

div.del_sym {
    width: 1.8ex;
    height: 1.8ex;
    background-image: url("data:image/svg+xml;utf8,%3Csvg viewBox='0 0 10 10' xmlns='http://www.w3.org/2000/svg'%3E%3Ccircle cx='5' cy='5' r='4.5' stroke='black' fill='none'/%3E%3Cpath d='M 2.88 2.88 L 7.12 7.12 M 7.12 2.88 L 2.88 7.12' stroke='black'/%3E%3C/svg%3E");
    background-size: cover;
}


/* Things for the question part */
.question {
    border: 1px solid black;
    border-radius: 3px;
    padding: 5px;
    background-color: #f3f3f3;
    margin-bottom: 10px;
    margin-top: 10px;
    margin-right: 40px;
    font-size: 100%;
    position: relative;
}

.fake_question {
    padding: 5px;
    margin-bottom: 20px;
    margin-top: 10px;
    margin-right: 40px;
    font-size: 100%;
    position: relative;
}

.blur {
    filter: blur(1px) opacity(70%);
}

.qro {
    transform: scale(0.8);
    margin-right: 0px;
    margin-top: 0px;
    margin-bottom: 0px;
}

.qro_container {
    position: relative;
    margin-right: 20px;
    margin-top: 10px;
    margin-bottom: 10px;
}

.question input {
    font-size: 100%;
}

.qtitle, .qtype, .qans {
    font-size: 120%;
    font-weight: bold;
    display: block;
    margin-bottom: 3px;
    margin-top: 5px;
}

.qtit {
    width: 80%;
}

.qtype, .qans {
    font-size: 110%;
}

.expand_sort > div, .expand_grade > div, .expand_select > div {
    margin-top: 5px;
    padding: 5px;
    width: 70%;
    border: 1px solid black;
    border-radius: 3px;
    background-color: #f8f8f8;
}

.expand_select > div {
    margin-left: 10px;
    margin-right: auto;
    width: 50%;
}

.expand_select input[type=number] {
    width: 3em;
    margin-top: 2px;
    margin-bottom: 2px;
}

.expand_sort > div {
    margin-left: auto;
    margin-right: auto;
}

.expand_grade > div {
    margin-left: auto;
    margin-right: 5px;
    width: 50%;
}

.expand_sort input[type=number] {
    width: 3em;
}

.expand_grade input, .answers input {
    width: 80%;
}

.mention, .fake_mention, .answer, .fake_answer, .new_trustee {
    width: 100%;
    margin-left: 5%;
    position: relative;
}

.answers {
    width: 80%;
}

.answer {
    margin-left: 5px;
}
.fake_answer {
    margin-left: 5px;
    margin-bottom: 6px;
}

.new_trustee {
    margin-left: 10%;
    margin-bottom: 6px;
}

.fake_mention {
    margin-top: 6px;
}

.fake_mention .d_i_side, .fake_answer .d_i_side, .new_trustee .d_i_side {
    top: 0px;
}

.fake_mention > div:first-child, .fake_answer > div:first-child, .new_trustee > div:first-child{
    width: 80%;
    text-align: right;
    font-style: italic;
    font-size: 90%;
}

.fake_question > div:first-child {
    text-align: right;
    font-style: italic;
    font-size: 110%;
}

.d_i_side {
    display: inline-block;
    position: absolute;
    top: 5px;
    left: calc(80% + 15px);
    font-size: 100%;
    font-weight: bold;
}

.fake_question .d_i_side {
    font-size: 130%;
    left: calc(100% + 10px);
}

.d_i_side div {
    display: inline-block;
    margin-right: 5px;
}


.d_i_side_top {
    display: inline-block;
    position: absolute;
    top: 0px;
    right: -50px;
    font-size: 130%;
    font-weight: bold;
}

.d_i_side_top div {
    display: inline-block;
    margin-right: 5px;
}

.mention .d_i_side, .answer .d_i_side {
    display: inline-block;
}


/* lists elections */
.answers_list {
    width: 80%;
    margin-top: 10px;
}
.answers_list input {
    width: 80%;
}
.answers_list .answer_label {
    width: 20%;
    display: inline-block;
}
.answers_list .answer_list_name input {
    width: 60%;
}

#previewbooth {
    text-align: center;
}

#previewbooth div {
    display: inline-block;
    border: 2px solid #ccc;
    border-radius: 10px;
    padding: 5px;
    background-color: #eee;
    font-size: 120%;
    margin-top: 20px;
    margin-bottom: 20px;
}

#previewbooth div:hover {
    background-color: #ccc;
    border-color: #aaa;
}

/* Things for the voters list part */

.main-zone__content > hr {
    margin-bottom: 20px;
}

#list_warning {
    text-align: center;
    font-size: 120%;
}
#list_warning strong {
    font-weight: bold;
}

#list_warning2 {
    text-align: center;
    font-style: italic;
    width: 80%;
    margin-left: auto;
    margin-right: auto;
}

.iflog, .ifw, .ifsetup, .ifopen {
    display: none;
}

#login:checked ~ table .iflog, #poids:checked ~ table .ifw, #rad_setup:checked ~ table .ifsetup, #rad_run:checked ~ table .ifopen {
    display: table-cell;
}

#rad_setup:checked ~ div.ifsetup {
    display: block;
}

#rad_run:checked ~ div.ifopen {
    display: block;
}

th {
    color: white;
    background-color: #49494b;
    padding: 10px;
    text-align: left;
}
th:first-child {
    width: 50%;
}
th:last-child {
    width: 10%;
    background-color: white;
}
.ifw {
    width: 10%;
}
.iflog {
    width: 30%;
}

td {
    border-bottom: 1px solid black;
    padding-left: 10px;
}
td:last-child {
    border-bottom: 0;
}

table {
    margin-top: 20px;
    margin-bottom: 20px;
    border-collapse: collapse;
    width: 100%;
}

#addtolist {
    margin-top: 20px;
}

#addtolist textarea {
    vertical-align:top;
}

.tooltip {
    position: relative;
    display: inline-block;
}

.tooltiptext {
    position:absolute;
    right:3ex;
    top: -3ex;
    line-height:2.5ex;
    border:1px solid black;
    font-size:90%;
    display:none;
    width:400px;
    padding:3px;
    background-color:#eee;
}

.tooltip:hover .tooltiptext {
    display:inline-block;
}

.tooltip div:first-child {
    display:inline-block;
    width:3ex;
    line-height:3ex;
    text-align:center;
    border-radius:1.5ex;
    background-color:black;
    color:white;
}

.trustee-link {
    border: solid 2px;
    border-radius: 0.5em;
    border-color: black;
    white-space: nowrap;
}

.trustee-link:hover {
    background-color: black !important;
    color: white !important;
}

.trustee-links {
    line-height: 3ex;
}
