/* UI Components try to respect [BEM](http://getbem.com/naming/) for the naming of their CSS classes */
@import "./components/PageHeader.css";
@import "./components/Progress.css";
@import "./components/InputCredentialSection.css";
@import "./components/CandidateWithCheckbox.css";
@import "./components/ClassicVoteCandidatesList.css";
@import "./components/QuestionWithVotableAnswers.css";
@import "./components/PageFooter.css";
@import "./components/VoteNavigation.css";
@import "./components/NiceButton.css";
@import "./components/NiceInput.css";
@import "./components/LoadingSpinner.css";
@import "./components/WholeVoteRecap.css";
@import "./components/ReviewEncryptSection.css";
@import "./components/MajorityJudgmentVoteBigCandidatesList.css";
@import "./components/MajorityJudgmentVoteSmallCandidatesList.css";
@import "./components/PreferentialVotingCandidatesList.css";
@import "./components/PreferentialVotingWithoutEqualityColumn.css";
@import "./components/CandidateList.css";
@import "./components/ListsVoteCandidatesList.css";
@import "./select-css.css";

body {
  background: var(--body-background);
  font-family: Arial, Helvetica, sans-serif;
  font-weight: lighter;
  line-height: initial;
}

h1,
h2,
h3 {
  font-family: "Ubuntu", arial, serif;
}

@media screen and (max-width: 640px) {
  body {
    margin: 0;
  }
}

@media screen and (min-width: 640px) {
  body {
    margin: 8px;
  }
}

input[type="password"],
input[type="text"] {
  font-size: 16px; /* this prevents Iphone browser from zooming in */
}

@supports (-webkit-touch-callout: none) {
  select {
    font-size: 16px; /* this prevents Iphone browser from zooming in */
  }
}

.page {
  max-width: 800px;
  margin: 0 auto;
  box-shadow: var(--page-shadow);
  border-radius: 10px;
}

.page-body {
  background: var(--background-light);
}

.markup-b {
  font-weight: bold;
}
.markup-i {
  font-style: italic;
}
.markup-error {
  color: var(--alert-text);
}

.current_step {
  text-align: center;
  font-size: 28px;
  padding-bottom: 28px;
}
